"""
Created on 3 Oct 2020

Static method tests for pymandel

@author: semuadmin
"""

import unittest

from pymandel.mandelbrot import hsv_to_rgb


class StaticTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testhsv2rgb(self):
        res = hsv_to_rgb(0.5, 0.2, 0.9)
        self.assertEqual(res, (183, 229, 229))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



"""
Mandelbrot generator - Main tkinter application class

Created on 29 Mar 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

import os
from tkinter import N, S, E, W, PhotoImage

from pymandel.about_dialog import AboutDialog
from pymandel.fractal_frame import FractalFrame
from pymandel.howto_dialog import HowtoDialog
from pymandel.menu_bar import MenuBar
from pymandel.settings_frame import SettingsFrame
from pymandel.status_frame import StatusFrame
from pymandel.strings import (
    INTROTXT,
    MENUHIDESE,
    MENUSHOWSE,
    MENUHIDESB,
    MENUSHOWSB,
    MENUSHOWAX,
    MENUHIDEAX,
    JITTXT,
)
from ._version import __version__

VERSION = __version__
DIRNAME = os.path.dirname(__file__)
ICON = os.path.join(DIRNAME, "resources/pymandel.png")


class App:
    """
    Main GUI Application Class
    """

    def __init__(self, master):
        """
        Set up main application and add frames
        """

        self.__master = master
        self.__master.columnconfigure(0, weight=1)
        self.__master.rowconfigure(0, weight=1)
        self.__master.protocol("WM_DELETE_WINDOW", self.exit)
        self.__master.title("PyMandel")
        self._show_settings = True  # Flag to toggle settings frame
        self._show_status = True  # Flag to toggle status bar
        self._show_axes = False  # Flag to toggle plot axes
        self.__master.iconphoto(True, PhotoImage(file=ICON))

        self.body()

    def body(self):
        """
        Set up frame and widgets
        """

        self.frm_status = StatusFrame(self, borderwidth=2, relief="groove")
        self.frm_settings = SettingsFrame(self, borderwidth=2, relief="groove")
        self.frm_fractal = FractalFrame(self, borderwidth=2, relief="groove")
        self.frm_fractal.grid(column=0, row=0, padx=2, pady=2, sticky=(N, S, E, W))
        self.frm_settings.grid(column=1, row=0, padx=2, pady=2, sticky=N)
        self.frm_status.grid(
            column=0, row=1, padx=2, pady=2, columnspan=2, sticky=(W, E)
        )

        self.menu = MenuBar(self)
        self.__master.config(menu=self.menu)

        # Set up keyboard hot keys
        self.__master.bind_all("<Control-q>", self.exit)

        self.set_status(JITTXT, "blue")
        self.frm_fractal.can_fractal.update()
        self.frm_fractal.plot()
        self.set_status(INTROTXT, "green")
        self.frm_fractal.focus_set()

    def toggle_settings(self):
        """
        Toggle Settings Frame on or off
        """

        if self._show_settings:
            self.frm_settings.grid_forget()
            self._show_settings = False
            self.menu.option_menu.entryconfig(5, label=MENUSHOWSE)
        else:
            self.frm_settings.grid(column=1, row=0, padx=2, pady=2, sticky=N)
            self._show_settings = True
            self.menu.option_menu.entryconfig(5, label=MENUHIDESE)

    def toggle_status(self):
        """
        Toggle Status Bar on or off
        """

        if self._show_status:
            self.frm_status.grid_forget()
            self._show_status = False
            self.menu.option_menu.entryconfig(6, label=MENUSHOWSB)
        else:
            self.frm_status.grid(column=0, row=1, columnspan=2, sticky=(W, E))
            self._show_status = True
            self.menu.option_menu.entryconfig(6, label=MENUHIDESB)

    def toggle_axes(self):
        """
        Toggle plot axes on or off
        """

        if self._show_axes:
            self.frm_fractal.show_axes = False
            self._show_axes = False
            self.menu.option_menu.entryconfig(7, label=MENUSHOWAX)
        else:
            self.frm_fractal.show_axes = True
            self._show_axes = True
            self.menu.option_menu.entryconfig(7, label=MENUHIDEAX)

    def set_status(self, message, color="black"):
        """
        Sets text of status bar
        """

        self.frm_status.set_status(message, color)

    def get_master(self):
        """
        Returns application master (Tk)
        """

        return self.__master

    def howto(self):
        """
        Open Howto dialog
        """

        HowtoDialog(self)

    def about(self):
        """
        Open About dialog
        """

        AboutDialog(self)

    def exit(self, *args, **kwargs):
        """
        Kill any running processes and quit application
        """

        self.frm_fractal.cancel_press()
        self.__master.destroy()



"""
Littéraux de chaîne de langue FRANCAIS pour l'application PyMandel tkinter

Créé le 22 avr 2020

@author: semuadmin
"""
# pylint: disable=line-too-long

WIKIURL = "https://fr.wikipedia.org/wiki/Ensemble_de_Mandelbrot"
GITHUBURL = "https://github.com/semuconsulting/PyMandel"
CETURL = "https://github.com/holoviz/colorcet/blob/master/LICENSE.txt"
MODULENAME = "pymandel"

COPYRIGHTTXT = "\u00A9 SEMU Consulting 2020\nGPLv3 License. Tous droits réservés"

COLORCETTXT = "Cartes de couleurs HoloViz Colorcet disponibles sous la licence\nCreative Commons Attribution (CC_BY)"

INTROTXT = "Bienvenue à PyMandel! Utilisez la molette de la souris ou un clic gauche pour zoomer, un clic droit pour centrer."

HELPTXT = (
    "Entrez les paramètres manuellement (ou importez-les à partir d'un fichier de métadonnées) et cliquez sur TRACER pour créer une image fractale avec les paramètres spécifiés.\n\n"
    + "Utilisez la molette de la souris pour effectuer un zoom avant et arrière à l'emplacement actuel du curseur.\n"
    + "Faites un clic gauche, faites glisser et relâchez - zoomez sur une zone rectangulaire dessinée.\n"
    + "Clic gauche - zoomer à l'emplacement du curseur de la valeur de l'incrément de zoom.\n"
    + "Maj et Clic gauche - zoom arrière.\n"
    + "Appuyez sur Gauche \u25C0 ou Droite \u25B6 en mode Julia pour faire pivoter l'ensemble Julia autour de son origine.\n\n"
    + "Bouton TRACER - trace l'image en utilisant les paramètres actuels.\n"
    + "Bouton Annuler - annule l'opération de tracé en cours.\n"
    + "Bouton de réinitialisation - réinitialise les paramètres aux valeurs par défaut.\n"
    + "Bouton Enregistrer - enregistre l'image actuellement affichée sous forme de fichier .png avec ses métadonnées associées sous forme de fichier .json.\n\n"
    + "Bouton Zoom - crée automatiquement une séquence d'images zoomées.\n"
    + "Bouton de rotation - crée automatiquement une séquence d'images Julia en rotation.\n\n"
    + "Fichier..Paramètres d'exportation - pour exporter les paramètres actuels (métadonnées).\n"
    + "Fichier..Paramètres d'importation - importer les métadonnées précédemment enregistrées.\n"
    + "Options..Masquer / Afficher les paramètres - active ou désactive le panneau des paramètres.\n"
    + "Options..Masquer / Afficher l'état - active ou désactive la barre d'état.\n"
    + "Aide..Comment - afficher cette boîte de dialogue Comment faire.\n"
    + "Aide..À propos - afficher la boîte de dialogue À propos."
)

ABOUTTXT = (
    "PyMandel est une application GUI open-source gratuite entièrement écrite en Python et tkinter avec des améliorations de performances Numba.\n\n"
    + "Les instructions et le code source sont disponibles sur Github au lien ci-dessous."
)

# Texte du message
JITTXT = "PREMIÈRE UTILISATION UNIQUEMENT: veuillez attendre la compilation et la mise en cache JIT"
SETINITTXT = "Paramètres initialisés"
VALERROR = "ERREUR! Veuillez corriger les entrées en surbrillance"
SAVEERROR = "ERREUR! Le fichier n'a pas pu être enregistré dans le répertoire spécifié"
METASAVEERROR = "ERREUR! Le fichier de métadonnées n'a pas pu être enregistré dans le répertoire spécifié"
NOIMGERROR = "ERREUR! Une image doit être créée avant l'enregistrement"
OPENFILEERROR = "ERREUR! Impossible d'ouvrir le fichier"
BADJSONERROR = "ERREUR! Fichier de métadonnées non valide"
SAVETITLE = "Sélectionner le répertoire d'enregistrement"
SELTITLE = "Sélectionner le fichier à importer"
METAPROMPTTXT = "importé, cliquez sur TRACER pour continuer"
IMGSAVETXT = "Image enregistrée sous"
COMPLETETXT = "Opération terminée en"
INPROGTXT = "Opération en cours ..."
OPCANTXT = "Opération annulée"
COORDTXT = "Coordonnées:"
COORDPOLTXT = "Coordonnées polaires:"
FRMTXT = "Cadre"
FRMSTXT = "Cadres"

# Texte du menu
MENUFILE = "Fichier"
MENUOPTIONS = "Options"
MENUSAVE = "Enregistrer l'image"
MENUEXPORT = "Exporter les paramètres"
MENUIMPORT = "Importer les paramètres"
MENUEXIT = "Quitter"
MENUPLOT = "TRACER Image"
MENUZOOM = "TRACER Zoom Animation"
MENUSPIN = "TRACER Julia Spin Animation"
MENUCAN = "Annuler"
MENURST = "Réinitialiser"
MENUHIDESE = "Masquer les paramètres"
MENUSHOWSE = "Afficher les paramètres"
MENUHIDESB = "Masquer la barre d'état"
MENUSHOWSB = "Afficher la barre d'état"
MENUHIDEAX = "Masquer les axes"
MENUSHOWAX = "Afficher les axes"
MENUHOWTO = "Comment faire"
MENUABOUT = "À propos"
MENUHELP = "Aide"

# Texte du bouton
BTNPLOT = "TRACER"
BTNSAVE = "Enregistrer"
BTNCAN = "Annuler"
BTNRST = "Réinitialiser"
BTNZOOM = "Zoom"
BTNSPIN = "Spin"

# Texte d'étiquette
LBLCTL = "Contrôles"
LBLSET = "Paramètres"
LBLMODE = "Définir le type"
LBLVAR = "Variant"
LBLAUTO = "Animer:"
LBLTHEME = "Thème\nCouleur"
LBLEXP = "Exposant"
LBLRAD = "Escape\nRadius"
LBLSHIFT = "Thème\nMaj"
LBLITER = "Max\nItérations"
LBLZOOM = "Zoom"
LBLZOOMINC = "Zoom\nIncrément"
LBLZXOFF = "Décalage ZX"
LBLZYOFF = "Décalage ZY"
LBLCX = "Julia CX"
LBLCY = "Julia CY"

# Texte de dialogue
DLGABOUT = "À propos de PyMandel"
DLGHOWTO = "Comment utiliser PyMandel"



"""
Status Bar frame class for tkinter application.

This handles the status bar notifications.

Created on 7 Apr 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import Frame, Label, StringVar, W, E


class StatusFrame(Frame):
    """
    Frame inheritance class for _status bar.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Frame.__init__(self, self.__master, *args, **kwargs)
        self._status = StringVar()

        self.body()

    def body(self):
        """
        Set up frame and widgets.
        """

        self.lbl_status = Label(self, textvariable=self._status, anchor=W)
        self.lbl_status.grid(column=0, row=0, sticky=(W, E))

    def set_status(self, message, color="black"):
        """
        Sets text of status bar.
        """

        self.lbl_status.config(fg=color)
        self._status.set("  " + message)

    def clear_status(self):
        """
        Clears text of status bar.
        """

        self._status.set("")



"""
How To Dialog Box class for tkinter application.

Created on 19 Apr 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import Toplevel, Label, Button, LEFT
from webbrowser import open_new_tab

from pymandel.strings import HELPTXT, COPYRIGHTTXT, DLGHOWTO, GITHUBURL


class HowtoDialog:
    """
    How To dialog box class
    """

    def __init__(self, app):
        """
        Initialise Toplevel dialog
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        self._dialog = Toplevel()
        self._dialog.title = DLGHOWTO
        self._dialog.geometry(
            f"+{self.__master.winfo_rootx() + 50}+{self.__master.winfo_rooty() + 50}"
        )
        self._dialog.attributes("-topmost", "true")

        self.body()

    def body(self):
        """
        Set up widgets
        """

        # Create widgets
        self.lbl_title = Label(self._dialog, text=DLGHOWTO)
        self.lbl_title.config(font=("Verdana", 16))

        self.lbl_desc = Label(self._dialog, justify=LEFT, text=HELPTXT, wraplength=500)

        self.lbl_copyright = Label(
            self._dialog, text=COPYRIGHTTXT, fg="blue", cursor="hand2"
        )
        self.btn_ok = Button(self._dialog, text="OK", width=8, command=self.ok_press)

        # Arrange widgets
        self.lbl_title.grid(column=0, row=0, padx=5, pady=5)
        self.lbl_desc.grid(column=0, row=2, padx=5, pady=5)
        self.lbl_copyright.grid(column=0, row=3, padx=5, pady=5)
        self.btn_ok.grid(column=0, row=5, ipadx=3, ipady=3, padx=5, pady=5)

        # Bind commands and hotkeys
        self.lbl_copyright.bind("<Button-1>", lambda e: open_new_tab(GITHUBURL))
        self.btn_ok.bind("<Return>", self.ok_press)
        self.btn_ok.focus_set()

    def ok_press(self, *args, **kwargs):
        """
        Handle OK button press
        """

        self.__master.update_idletasks()
        self._dialog.destroy()



"""
Menubar class for tkinter application.

This handles the menu bar.

Created on 27 Mar 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import Menu

from pymandel.strings import (
    MENUFILE,
    MENUOPTIONS,
    MENUSAVE,
    MENUEXPORT,
    MENUIMPORT,
    MENUEXIT,
    MENUPLOT,
    MENUZOOM,
    MENUSPIN,
    MENUCAN,
    MENURST,
    MENUHIDESE,
    MENUHIDESB,
    MENUSHOWAX,
    MENUHOWTO,
    MENUABOUT,
    MENUHELP,
)


class MenuBar(Menu):
    """
    Menu inheritance class for menu bar.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Menu.__init__(self, self.__master, *args, **kwargs)

        # Create a pull-down menu for file operations
        self.file_menu = Menu(self, tearoff=False)
        self.file_menu.add_command(
            label=MENUSAVE, underline=1, command=self.__app.frm_settings.save_image
        )
        self.file_menu.add_command(
            label=MENUEXPORT, underline=1, command=self.__app.frm_settings.save_metadata
        )
        self.file_menu.add_command(
            label=MENUIMPORT,
            underline=1,
            command=self.__app.frm_settings.import_metadata,
        )
        self.file_menu.add_command(
            label=MENUEXIT, underline=1, accelerator="Ctrl-Q", command=self.__app.exit
        )
        self.add_cascade(menu=self.file_menu, label=MENUFILE)

        # Create a pull-down menu for options
        self.option_menu = Menu(self, tearoff=False)
        self.option_menu.add_command(
            label=MENUPLOT, underline=1, command=self.__app.frm_fractal.plot
        )
        self.option_menu.add_command(
            label=MENUZOOM, underline=1, command=self.__app.frm_fractal.animate_zoom
        )
        self.option_menu.add_command(
            label=MENUSPIN, underline=1, command=self.__app.frm_fractal.animate_spin
        )
        self.option_menu.add_command(
            label=MENUCAN, underline=1, command=self.__app.frm_fractal.cancel_press
        )
        self.option_menu.add_command(
            label=MENURST, underline=1, command=self.__app.frm_settings.reset
        )
        self.option_menu.add_command(
            label=MENUHIDESE, underline=1, command=self.__app.toggle_settings
        )
        self.option_menu.add_command(
            label=MENUHIDESB, underline=1, command=self.__app.toggle_status
        )
        self.option_menu.add_command(
            label=MENUSHOWAX, underline=1, command=self.__app.toggle_axes
        )
        self.add_cascade(menu=self.option_menu, label=MENUOPTIONS)

        # Create a pull-down menu for help operations
        self.help_menu = Menu(self, tearoff=False)
        self.help_menu.add_command(
            label=MENUHOWTO, underline=1, command=self.__app.howto
        )
        self.help_menu.add_command(
            label=MENUABOUT, underline=1, command=self.__app.about
        )
        self.add_cascade(menu=self.help_menu, label=MENUHELP)



"""
DEUTSCHE Sprachzeichenfolgenliterale für die PyMandel tkinter-Anwendung

Erstellt am 22. April 2020

@author: semuadmin
"""
# pylint: disable=line-too-long

WIKIURL = "https://de.wikipedia.org/wiki/Mandelbrot-Menge"
GITHUBURL = "https://github.com/semuconsulting/PyMandel"
CETURL = "https://github.com/holoviz/colorcet/blob/master/LICENSE.txt"
MODULENAME = "pymandel"

COPYRIGHTTXT = "\u00A9 SEMU Consulting 2020\nGPLv3 lizenz. Alle Rechte vorbehalten"

COLORCETTXT = "HoloViz Colorcet-Farbkarten verfügbar unter der Lizenz\nCreative Commons Attribution (CC_BY)"

INTROTXT = "Willkommen bei PyMandel! Verwenden Sie das Mausrad oder klicken Sie mit der linken Maustaste, um zu zoomen, und klicken Sie mit der rechten Maustaste, um zu zentrieren."

HELPTXT = (
    "Geben Sie die Einstellungen manuell ein (oder importieren Sie sie aus einer Metadatendatei) und klicken Sie auf ZEICHNEN, um ein Fraktalbild mit den angegebenen Parametern zu erstellen.\n\n"
    + "Verwenden Sie das Mausrad, um an der aktuellen Cursorposition hinein- und herauszuzoomen.\n"
    + "Klicken Sie mit der linken Maustaste, ziehen Sie sie und lassen Sie sie los - zoomen Sie in einen gezeichneten rechteckigen Bereich.\n"
    + "Linksklick - Vergrößern Sie die Cursorposition um den Betrag des Zoominkrements.\n"
    + "Umschalt & Linksklick - Verkleinern.\n"
    + "Drücken Sie im Julia-Modus nach links \u25C0 oder rechts \u25B6, um das Julia-Set um seinen Ursprung zu drehen.\n\n"
    + "ZEICHNEN-Taste - Plotten Sie das Bild mit den aktuellen Einstellungen.\n"
    + "Abbrechen-Schaltfläche - Bricht den aktuellen Plotvorgang ab.\n"
    + "Reset-Taste - Setzt die Parameter auf die Standardwerte zurück.\n"
    + "Speichern - Speichern Sie das aktuell angezeigte Bild als PNG-Datei zusammen mit den zugehörigen Metadaten als JSON-Datei.\n\n"
    + "Zoom-Schaltfläche - Erstellt automatisch eine Folge von Zoom-Bildern.\n"
    + "Drehknopf - Erstellt automatisch eine Folge von 'sich drehenden' Julia-Bildern.\n\n"
    + "Datei..Exporteinstellungen - Aktuelle Einstellungen (Metadaten) exportieren.\n"
    + "Datei..Importeinstellungen - Importieren Sie zuvor gespeicherte Metadaten.\n"
    + "Optionen..Einstellungen ein- / ausblenden - Schaltet das Einstellungsfeld ein oder aus.\n"
    + "Optionen..Status ein- / ausblenden - Schaltet die Statusleiste ein oder aus.\n"
    + "Hilfe..Anleitung - Zeigen Sie diesen Dialog an.\n"
    + "Hilfe..Über - Dialogfeld Info anzeigen"
)

ABOUTTXT = (
    "PyMandel ist eine kostenlose Open-Source-GUI-Anwendung, die vollständig in Python und tkinter mit Numba-Leistungsverbesserungen geschrieben wurde.\n\n"
    + "Anweisungen und Quellcode sind auf Github unter dem folgenden Link verfügbar."
)

# Nachrichtentext
JITTXT = "NUR FÜR DEN ERSTEN GEBRAUCH: Bitte warten Sie auf JIT Compilation and Caching"
SETINITTXT = "Einstellungen initialisiert"
VALERROR = "FEHLER! Bitte hervorgehobene Einträge korrigieren"
SAVEERROR = "FEHLER! Datei konnte nicht im angegebenen Verzeichnis gespeichert werden"
METASAVEERROR = (
    "FEHLER! Metadatendatei konnte nicht im angegebenen Verzeichnis gespeichert werden"
)
NOIMGERROR = "FEHLER! Vor dem Speichern muss ein Bild erstellt werden."
OPENFILEERROR = "FEHLER! Datei konnte nicht geöffnet werden"
BADJSONERROR = "FEHLER! Ungültige Metadatendatei"
SAVETITLE = "Verzeichnis speichern auswählen"
SELTITLE = "Datei zum Importieren auswählen"
METAPROMPTTXT = "importiert, klicken Sie auf ZEICHNEN, um fortzufahren"
IMGSAVETXT = "Bild gespeichert als"
COMPLETETXT = "Vorgang abgeschlossen in"
INPROGTXT = "Operation läuft ..."
OPCANTXT = "Vorgang abgebrochen"
COORDTXT = "Koordinaten:"
COORDPOLTXT = "Polarkoordinaten:"
FRMTXT = "Rahmen"
FRMSTXT = "Rahmen"

# Menütext
MENUFILE = "Datei"
MENUOPTIONS = "Optionen"
MENUSAVE = "Bild speichern"
MENUEXPORT = "Exporteinstellungen"
MENUIMPORT = "Importeinstellungen"
MENUEXIT = "Beenden"
MENUPLOT = "Bild zeichnen"
MENUZOOM = "Zeichnen Zoom Animation"
MENUSPIN = "Zeichnen Julia Rotieren Animation"
MENUCAN = "Abbrechen"
MENURST = "Zurücksetzen"
MENUHIDESE = "Einstellungen ausblenden"
MENUSHOWSE = "Einstellungen anzeigen"
MENUHIDESB = "Statusleiste ausblenden"
MENUSHOWSB = "Statusleiste anzeigen"
MENUHIDEAX = "Achsen ausblenden"
MENUSHOWAX = "Achsen anzeigen"
MENUHOWTO = "Verwendun"
MENUABOUT = "Über"
MENUHELP = "Hilfe"

# Schaltflächentext
BTNPLOT = "ZEICHNEN"
BTNSAVE = "Speichern"
BTNCAN = "Abbrechen"
BTNRST = "Zurücksetzen"
BTNZOOM = "Zoom"
BTNSPIN = "Rotieren"

# Beschriftungstext
LBLCTL = "Kontrollen"
LBLSET = "Einstellungen"
LBLMODE = "Serietyp"
LBLVAR = "Variant"
LBLAUTO = "Animieren:"
LBLTHEME = "Thema"
LBLEXP = "Exponent"
LBLRAD = "Escape\nRadius"
LBLSHIFT = "Thema\nVerschiebung"
LBLITER = "Max\nIterationen"
LBLZOOM = "Zoom"
LBLZOOMINC = "Zoom\nInkrement"
LBLZXOFF = "ZX Versatz"
LBLZYOFF = "ZY Versatz"
LBLCX = "Julia CX"
LBLCY = "Julia CY"

# Dialogtext
DLGABOUT = "Über PyMandel"
DLGHOWTO = "Verwendung von PyMandel"



"""
Fractal frame class for tkinter application.

This handles all the user interaction (mouse and key clicks) and image mechanics.

Created on 3 Apr 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from cmath import polar
from math import sqrt, log, pi, sin, cos
from platform import system
from time import time
from tkinter import Frame, Canvas, NW, BOTH, YES

from PIL import ImageTk

from pymandel.mandelbrot import (
    Mandelbrot,
    ptoc,
    ctop,
    MODES,
    VARIANTS,
    MANDELBROT,
    JULIA,
    TRICORN,
    BURNINGSHIP,
    STANDARD,
)
from pymandel.strings import (
    COMPLETETXT,
    INPROGTXT,
    OPCANTXT,
    SAVEERROR,
    COORDTXT,
    FRMTXT,
)

ZOOM = 0
SPIN = 1
ZOOMOUT = 0
ZOOMIN = 1
GOJULIA = 2


class FractalFrame(Frame):
    """
    Frame inheritance class for plotting area.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        self._fractal = None  # Must be instance variable to persist after use
        self._animating = False
        self._setmode = MANDELBROT
        self._setvar = STANDARD
        self._leftclickmode = ZOOMIN
        self.show_axes = False
        self._zoom_rect = None
        self._x_start = None
        self._y_start = None
        self._xaxis = self._yaxis = 0

        self.mandelbrot = None

        self.body()

    def body(self):
        """
        Set up frame and widgets.
        """

        self.__master.update_idletasks()
        plot_height = self.__app.frm_settings.winfo_reqheight() - 8
        self.can_fractal = Canvas(
            self, width=plot_height * 1.5, height=plot_height, cursor="tcross"
        )
        self.can_fractal.pack(fill=BOTH, expand=YES)
        self.can_fractal.bind("<Motion>", self.get_coords)
        self.can_fractal.bind("<Button-1>", self.on_left_click)  # Left-click zoom
        self.can_fractal.bind("<Button-3>", self.on_right_click)  # Right-click center
        self.can_fractal.bind(
            "<Button-2>", self.on_right_click
        )  # Right-click center (MacOS)
        self.can_fractal.bind_all("<KeyPress>", self.on_key_down)
        self.can_fractal.bind_all("<KeyRelease>", self.on_key_release)
        self.can_fractal.bind("<MouseWheel>", self.on_mouse_wheel)  # Mousewheel zoom
        self.can_fractal.bind("<ButtonPress-1>", self.on_button_down)  # Left-down
        self.can_fractal.bind("<B1-Motion>", self.on_button_drag)  # Left-drag
        self.can_fractal.bind(
            "<ButtonRelease-1>", self.on_button_release
        )  # Left-release

    def plot(self):
        """
        Plot Mandelbrot set as an ImageTk.PhotoImage and load
        this into the GUI's Canvas widget for display.
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings.get("valid"):
            return

        # Apply the current settings
        width, height = self.get_size()
        setmode = settings.get("settype")
        if setmode == "Julia":
            self._setmode = JULIA
        else:
            self._setmode = MANDELBROT
        setvar = settings.get("setvar")
        if setvar == VARIANTS[TRICORN]:
            self._setvar = TRICORN
        elif setvar == VARIANTS[BURNINGSHIP]:
            self._setvar = BURNINGSHIP
        else:
            self._setvar = STANDARD
        zoom = settings.get("zoom")
        radius = settings.get("radius")
        exponent = settings.get("exponent")
        if settings.get("autoiter"):
            maxiter = self.get_autoiter(zoom)
            self.__app.frm_settings.update_settings(maxiter=maxiter)
        else:
            maxiter = settings.get("maxiter")
        zx_off = settings.get("zxoffset")
        zy_off = settings.get("zyoffset")
        cx_off = settings.get("cxoffset")
        cy_off = settings.get("cyoffset")
        theme = settings.get("theme")
        shift = settings.get("shift")

        if not self._animating:
            self.__app.set_status(INPROGTXT)
        self.__master.update_idletasks()
        start = time()

        self.mandelbrot = Mandelbrot(self)
        self.mandelbrot.plot_image(
            self._setmode,
            self._setvar,
            width,
            height,
            zoom,
            radius,
            exponent,
            zx_off,
            zy_off,
            maxiter,
            theme,
            shift,
            cx_off,
            cy_off,
        )
        self._fractal = ImageTk.PhotoImage(self.mandelbrot.get_image())
        self.can_fractal.create_image(
            0, 0, image=self._fractal, state="normal", anchor=NW
        )

        if self.show_axes:
            self.axes(width, height)
        self.can_fractal.update()

        if (
            not self.mandelbrot.get_cancel() and not self._animating
        ):  # If plot wasn't cancelled
            end = time()
            self.__app.set_status(COMPLETETXT + str(round(end - start, 2)) + " seconds")

    def axes(self, width, height):
        """
        Draw complex space axes on plot.
        Bit simple - only really works at low magnifications
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings.get("valid"):
            return

        zoom = settings.get("zoom")
        zxoff = settings.get("zxoffset")
        zyoff = settings.get("zyoffset")
        tki = [-2.5, -2, -1.5, -1, -0.5, 0.5, 1, 1.5, 2, 2.5]
        tick = 10
        xoff, yoff = ctop(width, height, 0, 0, zxoff, zyoff, zoom)
        xline = [0, height / 2 + yoff, width, height / 2 + yoff]
        yline = [width / 2 - xoff, 0, width / 2 - xoff, height]
        self.can_fractal.create_line(xline, fill="gray")
        self.can_fractal.create_text(
            0 + 40, height / 2 + yoff + 20, text="Re[c] (X)", fill="gray", anchor="n"
        )
        self.can_fractal.create_line(yline, fill="gray")
        self.can_fractal.create_text(
            width / 2 - xoff + 20, 0 + 20, text="Im[c] (Y)", fill="gray", anchor="w"
        )

        for x_axis in tki:
            xoff, yoff = ctop(width, height, x_axis, 0, zxoff, zyoff, zoom)
            yline = [
                width / 2 - xoff,
                height / 2 - tick + yoff,
                width / 2 - xoff,
                height / 2 + tick + yoff,
            ]
            self.can_fractal.create_line(yline, fill="gray")
            self.can_fractal.create_text(
                width / 2 - xoff,
                height / 2 - tick + yoff - 10,
                text=x_axis * -1,
                fill="gray",
            )
        for y_axis in tki:
            xoff, yoff = ctop(width, height, 0, y_axis, zxoff, zyoff, zoom)
            xline = [
                width / 2 - tick - xoff,
                height / 2 + yoff,
                width / 2 + tick - xoff,
                height / 2 + yoff,
            ]
            self.can_fractal.create_line(xline, fill="gray")
            self.can_fractal.create_text(
                width / 2 - tick - xoff - 10,
                height / 2 + yoff,
                text=y_axis * -1,
                fill="gray",
                anchor="e",
            )
        self.can_fractal.update()

    def get_autoiter(self, zoom):
        """
        Arbitrary algorithm to derive 'optimal' max iterations from zoom level.
        """

        settype = self.__app.frm_settings.get_settings().get("settype")
        if settype == MODES[JULIA]:
            miniter = 500
        else:
            miniter = 100
        maxiter = max(miniter, int(abs(1000 * log(1 / sqrt(zoom)))))
        return maxiter

    def cancel_press(self):
        """
        Cancel in-progress plot.
        """

        self.mandelbrot.cancel_plot()
        self._animating = False
        self.__app.set_status(OPCANTXT, "red")

    def get_size(self):
        """
        Get current canvas size.
        """

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.can_fractal.winfo_width()
        height = self.can_fractal.winfo_height()
        return (width, height)

    def get_coords(self, event):
        """
        Get and display complex space coordinates of current mouse location.
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings.get("valid"):
            return (0, 0)

        # Get actual screen dimensions and mouse location
        width, height = self.get_size()
        x_pos = event.x
        y_pos = event.y
        zoom = settings.get("zoom")
        zxoff = settings.get("zxoffset")
        zyoff = settings.get("zyoffset")

        # Convert to complex space coordinates (offsets are expressed in z space)
        zx_coord, zy_coord = ptoc(width, height, x_pos, y_pos, zxoff, zyoff, zoom)

        self.__app.set_status(
            COORDTXT + " Re[c] (X): " + str(zx_coord) + ",  Im[c] (Y): " + str(zy_coord)
        )
        return (zx_coord, zy_coord)

    def on_mouse_wheel(self, event):
        """
        Zoom in and out using mouse wheel

        Mouse wheel event.delta increment differs between Windows and
        MacOS/Linux platforms. These settings seem to work OK for Logitech
        and Apple wireless mice but may need to be tweaked for other devices
        """

        if system() == "Windows":
            sensitivity = 0.015
        else:
            sensitivity = 1.5
        self.cancel_press()
        zx_coord, zy_coord = self.get_coords(event)
        zoom = self.__app.frm_settings.get_settings().get("zoom")

        if event.delta > 0:
            zoom = zoom * (event.delta * sensitivity)
        else:
            zoom = zoom / (event.delta * -sensitivity)

        self.__app.frm_settings.update_settings(
            zoom=zoom, zxoffset=zx_coord, zyoffset=zy_coord
        )
        self.__master.update_idletasks()

        self.plot()

    def on_right_click(self, event):
        """
        Right-Click - center and redraw image at mouse position.
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings.get("valid"):
            return

        self.cancel_press()
        zx_coord, zy_coord = self.get_coords(event)
        self.__app.frm_settings.update_settings(zxoffset=zx_coord, zyoffset=zy_coord)

        self.plot()

    def on_left_click(self, event):
        """
        Left-click - zoom in at cursor position (ZOOMIN mode).
        Left-click & Shift - zoom out (ZOOMOUT mode).
        Left-click & Ctrl-L or Alt-L - switch to Julia mode and plot Julia set
        corresponding to current cursor (cx, cy) position (GOJULIA mode).
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings.get("valid"):
            return

        self.cancel_press()
        zx_coord, zy_coord = self.get_coords(event)
        zoom = settings.get("zoom")
        zoominc = settings.get("zoominc")
        if self._leftclickmode == ZOOMIN:
            zoom = zoom * zoominc
            self.__app.frm_settings.update_settings(
                zoom=zoom, zxoffset=zx_coord, zyoffset=zy_coord
            )
        if self._leftclickmode == ZOOMOUT:
            zoom = zoom / zoominc
            self.__app.frm_settings.update_settings(
                zoom=zoom, zxoffset=zx_coord, zyoffset=zy_coord
            )
        if self._leftclickmode == GOJULIA:
            self.__app.frm_settings.update_settings(
                settype=MODES[JULIA],
                zxoffset=0,
                zyoffset=0,
                cxoffset=zx_coord,
                cyoffset=zy_coord,
            )

        self.plot()

    def on_key_down(self, event):
        """
        Left-click - zoom in at cursor position (ZOOMIN mode).
        Left-click & Shift - zoom out (ZOOMOUT mode).
        Left-click & Ctrl-L or Alt-L - switch to Julia mode and plot Julia set
        corresponding to current cursor (cx, cy) position (GOJULIA mode).
        Left or Right Arrow keys - in Julia mode, rotate Julia set about its origin

        (for some reason tkinter doesn't appear to recognise an Alt-L event on all
        Windows platforms)
        """

        if event.keysym == "Shift_L":
            self._leftclickmode = ZOOMOUT
            self.can_fractal.config(cursor="sizing")

        if event.keysym in ("Alt_L", "Control_L"):
            self._leftclickmode = GOJULIA
            self._setmode = JULIA
            self.can_fractal.config(cursor="target")

        # Pressing Left or Right button in Julia mode will rotate the Julia Set clockwise
        # or anti-clockwise about its origin
        if (event.keysym in ("Left", "Right")) and self._setmode == JULIA:
            if event.keysym == "Left":
                self.rotate_julia(0.01)
            else:
                self.rotate_julia(-0.01)

            self.plot()

    def on_key_release(self, event):
        """
        Revert to normal left-click mode (ZOOMIN)
        """

        self._leftclickmode = ZOOMIN
        self.can_fractal.config(cursor="tcross")

    def on_button_down(self, event):
        """
        Left-click - start to draw zoom area
        """

        self._x_start = event.x
        self._y_start = event.y
        if not self._zoom_rect:
            self._zoom_rect = self.can_fractal.create_rectangle(
                self._xaxis, self._yaxis, 1, 1, outline="yellow", width=2
            )

    def on_button_drag(self, event):
        """
        Left-drag - extend drawn zoom area
        """

        self.can_fractal.coords(
            self._zoom_rect, self._x_start, self._y_start, event.x, event.y
        )

    def on_button_release(self, event):
        """
        Left-release - zoom to drawn area
        """

        if not self._zoom_rect:
            return

        # Bug out if settings invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings["valid"]:
            return

        zxoff = settings.get("zxoffset")
        zyoff = settings.get("zyoffset")
        zoom = settings.get("zoom")
        width, height = self.get_size()
        x_0, y_0, x_1, y_1 = self.can_fractal.coords(self._zoom_rect)

        # If rectangle is less than two pixels wide, treat as momentary left-click
        # and simply zoom in by Zoom Increment amount
        if abs(x_1 - x_0) <= 2:
            self._zoom_rect = None
            self.on_left_click(event)
            return

        # Find the complex coordinates of the centre of the drawn rectangle
        x_pos = x_0 + ((x_1 - x_0) / 2)
        y_pos = y_0 + ((y_1 - y_0) / 2)
        zx_coord, zy_coord = ptoc(width, height, x_pos, y_pos, zxoff, zyoff, zoom)
        # Base zoom on x-axis only to maintain current canvas aspect ratio
        zoom = zoom * width / (x_1 - x_0)
        self.__app.frm_settings.update_settings(
            zoom=zoom, zxoffset=zx_coord, zyoffset=zy_coord
        )

        self.plot()
        self._zoom_rect = None

    def rotate_julia(self, angle):
        """
        Rotate Julia set about its origin by the angle in radians
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings["valid"]:
            return

        cx_off = settings.get("cxoffset")
        cy_off = settings.get("cyoffset")
        rds, phi = polar(complex(cx_off, cy_off))
        angle = phi + angle
        cx_off = rds * cos(angle)
        cy_off = rds * sin(angle)
        self.__app.frm_settings.update_settings(cxoffset=cx_off, cyoffset=cy_off)

    def animate_zoom(self):
        """
        Animate Julia spin mode
        """
        self.animate(ZOOM)

    def animate_spin(self):
        """
        Animate zoom mode
        """
        self.animate(SPIN)

    def animate(self, animatemode):
        """
        Generates and saves a series of frames at a specified offset.
        The individual frames can then be assembled into e.g. an
        animated .gif or short video using external tools.

        Also used to generate a 'Spinning Julia' animation, rotating
        a Julia Set anti-clockwise around its origin.
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings.get("valid"):
            return

        # If autosave is on, set filepath if not already set
        if settings.get("autosave"):
            filepath = self.__app.frm_settings.set_filepath()
            if filepath is None:  # User cancelled
                return

        zoom = settings.get("zoom")
        if settings.get("autoiter"):
            maxiter = self.get_autoiter(zoom)
            self.__app.frm_settings.update_settings(maxiter=maxiter)
        else:
            maxiter = settings.get("maxiter")
        frames = settings.get("frames")
        zoominc = settings.get("zoominc")
        name = settings.get("filename")
        width, height = self.get_size()

        self.mandelbrot = Mandelbrot(self)
        self.mandelbrot.cancel_plot()  # Cancel any in-flight plot
        self._animating = True

        start = time()
        for i in range(frames):
            self.__app.set_status(FRMTXT + " " + str(i) + " / " + str(frames) + " ...")
            self.can_fractal.update()

            if animatemode == SPIN:  # Spinning Julia animation
                self.rotate_julia((1 / frames) * 2 * pi)

            self.plot()

            if self.show_axes:
                self.axes(width, height)
            self.can_fractal.update()

            if self.mandelbrot.get_cancel():
                return

            if settings.get("autosave"):
                fqname = filepath + "/" + name + "_" + str(i + 1).zfill(3)
                try:
                    image = self.mandelbrot.get_image()
                    image.save(fqname + ".png", format="png")
                except OSError:
                    self.__app.set_status(SAVEERROR, "red")
                    self.__app.filepath = None
                    return

            if animatemode == ZOOM:
                zoom = zoom * zoominc
                maxiter = self.get_autoiter(zoom)
                self.__app.frm_settings.update_settings(zoom=zoom, maxiter=maxiter)

        end = time()
        self.__app.set_status(COMPLETETXT + str(round(end - start, 2)) + " seconds")
        self._animating = False



"""
Created on 27 Sep 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from pymandel._version import __version__

version = __version__



"""
Settings frame class for tkinter application.

This handles all the settings and file storage/retrieval.

Created on 3 Apr 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import (
    ttk,
    Frame,
    Button,
    Label,
    Entry,
    Checkbutton,
    Listbox,
    Spinbox,
    Scale,
    filedialog,
    Scrollbar,
    IntVar,
    DoubleVar,
    StringVar,
    N,
    S,
    E,
    W,
    LEFT,
    RIGHT,
    HORIZONTAL,
    VERTICAL,
    NORMAL,
    DISABLED,
)

from time import strftime, gmtime

from re import match

from json import dump, loads

import os

from pymandel.strings import (
    MODULENAME,
    SAVETITLE,
    VALERROR,
    SAVEERROR,
    SETINITTXT,
    METASAVEERROR,
    NOIMGERROR,
    OPENFILEERROR,
    BADJSONERROR,
    SELTITLE,
    METAPROMPTTXT,
    IMGSAVETXT,
    BTNPLOT,
    BTNSAVE,
    BTNCAN,
    BTNRST,
    BTNZOOM,
    BTNSPIN,
    LBLMODE,
    LBLVAR,
    LBLAUTO,
    LBLTHEME,
    LBLEXP,
    LBLRAD,
    LBLSHIFT,
    LBLITER,
    LBLZOOM,
    LBLZOOMINC,
    LBLZXOFF,
    LBLZYOFF,
    LBLCX,
    LBLCY,
    FRMSTXT,
)

from pymandel.mandelbrot import MODES, VARIANTS, THEMES, MANDELBROT, JULIA

# Global variables for Entry widget validation
GOOD = "azure"
BAD = "pink1"


class SettingsFrame(Frame):
    """
    Frame inheritance class for application settings and controls.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Frame.__init__(self, self.__master, *args, **kwargs)

        #  Initialise up key variables
        self._image_num = 0
        self._image_name = "image"
        self._settype = StringVar()
        self._setvar = StringVar()
        self._zoom = DoubleVar()
        self._radius = DoubleVar()
        self._exponent = DoubleVar()
        self._zx_off = DoubleVar()
        self._zy_off = DoubleVar()
        self._cx_off = DoubleVar()
        self._cy_off = DoubleVar()
        self._maxiter = IntVar()
        self._zx_coord = DoubleVar()
        self._zy_coord = DoubleVar()
        self._theme = StringVar()
        self._shift = IntVar()
        self._themes = None
        self._filename = StringVar()
        self._filepath = None
        self._frames = IntVar()
        self._zoominc = DoubleVar()
        self._autoiter = IntVar()
        self._autosave = IntVar()
        self._validsettings = True

        self.body()

    def body(self):
        """
        Set up frame and widgets.
        """

        # Create settings panel widgets
        # pylint: disable=W0108
        self.lbl_settype = Label(self, text=LBLMODE)
        self.spn_settype = Spinbox(
            self,
            values=MODES,
            width=8,
            bg=GOOD,
            wrap=True,
            textvariable=self._settype,
        )
        self.lbl_setvar = Label(self, text=LBLVAR)
        self.spn_setvar = Spinbox(
            self,
            values=VARIANTS,
            width=8,
            bg=GOOD,
            wrap=True,
            textvariable=self._setvar,
        )
        self.lbl_zoom = Label(self, text=LBLZOOM)
        self.ent_zoom = Entry(
            self,
            border=2,
            relief="sunken",
            width=12,
            bg=GOOD,
            justify=RIGHT,
            textvariable=self._zoom,
        )
        self.lbl_zoominc = Label(self, text=LBLZOOMINC, justify=LEFT)
        self.ent_zoominc = Entry(
            self, width=5, border=2, bg=GOOD, justify=RIGHT, textvariable=self._zoominc
        )
        self.lbl_zx_off = Label(self, text=LBLZXOFF)
        self.ent_zx_off = Entry(
            self,
            border=2,
            relief="sunken",
            width=12,
            bg=GOOD,
            justify=RIGHT,
            textvariable=self._zx_off,
        )
        self.lbl_zy_off = Label(self, text=LBLZYOFF)
        self.ent_zy_off = Entry(
            self,
            border=2,
            relief="sunken",
            width=12,
            bg=GOOD,
            justify=RIGHT,
            textvariable=self._zy_off,
        )
        self.lbl_cx_off = Label(self, text=LBLCX)
        self.ent_cx_off = Entry(
            self,
            border=2,
            relief="sunken",
            width=12,
            bg=GOOD,
            justify=RIGHT,
            textvariable=self._cx_off,
            state=DISABLED,
        )
        self.lbl_cy_off = Label(self, text=LBLCY)
        self.ent_cy_off = Entry(
            self,
            border=2,
            relief="sunken",
            width=12,
            bg=GOOD,
            justify=RIGHT,
            textvariable=self._cy_off,
            state=DISABLED,
        )
        self.lbl_niter = Label(self, text=LBLITER, justify=LEFT)
        self.ent_maxiter = Entry(
            self,
            border=2,
            relief="sunken",
            bg=GOOD,
            width=8,
            justify=RIGHT,
            textvariable=self._maxiter,
        )
        self.chk_autoiter = Checkbutton(
            self, text="Auto", variable=self._autoiter, onvalue=1, offvalue=0
        )
        self.lbl_theme = Label(self, text=LBLTHEME, justify=LEFT)
        self.lbl_radius = Label(self, text=LBLRAD, justify=LEFT)
        self.ent_radius = Entry(
            self,
            border=2,
            relief="sunken",
            bg=GOOD,
            width=8,
            justify=RIGHT,
            textvariable=self._radius,
        )
        self.lbl_exp = Label(self, text=LBLEXP)
        self.ent_exp = Entry(
            self,
            border=2,
            relief="sunken",
            bg=GOOD,
            width=8,
            justify=RIGHT,
            textvariable=self._exponent,
        )
        self.sep_1 = ttk.Separator(
            self,
            orient=HORIZONTAL,
        )
        self.lbx_theme = Listbox(
            self,
            border=2,
            relief="sunken",
            bg=GOOD,
            width=6,
            height=5,
            justify=LEFT,
            exportselection=False,
        )
        self.lbl_shift = Label(self, text=LBLSHIFT, justify=LEFT)
        self.scl_shift = Scale(
            self,
            from_=0,
            to=100,
            orient=HORIZONTAL,
            variable=self._shift,
            border=2,
            relief="sunken",
            sliderlength=20,
            troughcolor=GOOD,
        )
        self.scrollbar = Scrollbar(self, orient=VERTICAL)
        self.lbx_theme.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.lbx_theme.yview)
        self.lbx_theme.select_set(0)
        self.lbx_theme.event_generate("<<ListboxSelect>>")
        self.lbl_coords = Label(self, text="Re, Im", fg="grey")
        self.btn_plot = Button(
            self,
            text=BTNPLOT,
            width=8,
            fg="green",
            command=lambda: self.__app.frm_fractal.plot(),
        )
        self.btn_cancel = Button(
            self,
            text=BTNCAN,
            width=8,
            command=lambda: self.__app.frm_fractal.cancel_press(),
        )
        self.btn_reset = Button(self, text=BTNRST, width=8, command=self.reset)
        self.ent_save = Entry(
            self,
            textvariable=self._filename,
            width=6,
            border=2,
            relief="sunken",
            bg=GOOD,
            justify=LEFT,
        )
        self._filename.set(self._image_name + str(self._image_num))
        self.btn_save = Button(self, text=BTNSAVE, width=8, command=self.save_image)
        self.lbl_auto = Label(self, text=LBLAUTO, justify=LEFT)
        self.btn_autozoom = Button(
            self,
            text=BTNZOOM,
            width=8,
            command=lambda: self.__app.frm_fractal.animate_zoom(),
        )
        self.btn_autospin = Button(
            self,
            text=BTNSPIN,
            width=8,
            command=lambda: self.__app.frm_fractal.animate_spin(),
            state=DISABLED,
        )
        self.chk_autosave = Checkbutton(
            self, text=BTNSAVE, variable=self._autosave, onvalue=1, offvalue=0
        )
        self.lbl_frames = Label(self, text=FRMSTXT)
        self.ent_frames = Entry(
            self, width=5, border=2, bg=GOOD, justify=RIGHT, textvariable=self._frames
        )

        # Get list of available themes
        for idx, theme in enumerate(THEMES):
            self.lbx_theme.insert(idx, theme)

        self.body_arrange()  # Position all widgets in frame

        self.reset()  # Reset all settings to their defaults

        self.set_traces()  # Trace entry variables for validation and event handling

    def body_arrange(self):
        """
        Position widgets in frame
        """

        # Position all widgets in their parent frames
        self.btn_plot.grid(column=0, row=1, ipadx=3, ipady=3, sticky=W, padx=3, pady=3)
        self.btn_cancel.grid(
            column=1, row=1, ipadx=3, ipady=3, sticky=W, padx=3, pady=3
        )
        self.btn_reset.grid(column=2, row=1, ipadx=3, ipady=3, sticky=W, padx=3, pady=3)
        self.ent_save.grid(column=0, row=2, columnspan=2, sticky=(W, E), padx=3, pady=3)
        self.btn_save.grid(column=2, row=2, ipadx=3, ipady=3, sticky=W, padx=3, pady=3)
        self.lbl_auto.grid(column=0, row=3, sticky=W)
        self.btn_autozoom.grid(
            column=1, row=3, ipadx=3, ipady=3, sticky=W, padx=3, pady=3
        )
        self.btn_autospin.grid(
            column=2, row=3, ipadx=3, ipady=3, sticky=W, padx=3, pady=3
        )
        self.lbl_frames.grid(column=0, row=4, sticky=W)
        self.ent_frames.grid(column=1, row=4, sticky=W, padx=3, pady=3)
        self.chk_autosave.grid(column=2, row=4, sticky=W, padx=3, pady=3)
        self.sep_1.grid(column=0, row=5, columnspan=3, pady=5, sticky=(W, E))
        self.lbl_settype.grid(column=0, row=6, sticky=W)
        self.spn_settype.grid(
            column=1, row=6, columnspan=2, sticky=(W, E), padx=3, pady=3
        )
        self.lbl_setvar.grid(column=0, row=7, sticky=W)
        self.spn_setvar.grid(
            column=1, row=7, columnspan=2, sticky=(W, E), padx=3, pady=3
        )
        self.lbl_zoom.grid(column=0, row=8, sticky=W)
        self.ent_zoom.grid(column=1, row=8, columnspan=2, sticky=(W, E), padx=3, pady=3)
        self.lbl_zoominc.grid(column=0, row=9, sticky=W)
        self.ent_zoominc.grid(column=1, row=9, sticky=W, padx=3, pady=3)
        self.lbl_zx_off.grid(column=0, row=10, sticky=W)
        self.ent_zx_off.grid(
            column=1, row=10, columnspan=2, sticky=(W, E), padx=3, pady=3
        )
        self.lbl_zy_off.grid(column=0, row=11, sticky=W)
        self.ent_zy_off.grid(
            column=1, row=11, columnspan=2, sticky=(W, E), padx=3, pady=3
        )
        self.lbl_cx_off.grid(column=0, row=12, sticky=W)
        self.ent_cx_off.grid(
            column=1, row=12, columnspan=2, sticky=(W, E), padx=3, pady=3
        )
        self.lbl_cy_off.grid(column=0, row=13, sticky=W)
        self.ent_cy_off.grid(
            column=1, row=13, columnspan=2, sticky=(W, E), padx=3, pady=3
        )
        self.lbl_niter.grid(column=0, row=14, sticky=W)
        self.ent_maxiter.grid(column=1, row=14, sticky=W, padx=3, pady=3)
        self.chk_autoiter.grid(column=2, row=14, sticky=W, padx=3, pady=3)
        self.lbl_radius.grid(column=0, row=15, sticky=W)
        self.ent_radius.grid(column=1, row=15, sticky=W, padx=3, pady=3)
        self.lbl_exp.grid(column=0, row=16, sticky=W)
        self.ent_exp.grid(column=1, row=16, sticky=W, padx=3, pady=3)
        self.lbl_theme.grid(column=0, row=17, sticky=W)
        self.lbx_theme.grid(
            column=1, row=17, padx=3, pady=3, columnspan=2, sticky=(N, S, W, E)
        )
        self.scrollbar.grid(column=2, row=17, sticky=(N, S, E))
        self.lbl_shift.grid(column=0, row=18, sticky=W)
        self.scl_shift.grid(
            column=1, row=18, columnspan=2, padx=3, pady=3, sticky=(W, E)
        )
        self.lbx_theme.bind("<<ListboxSelect>>", self.get_sel_theme)

    def set_traces(self):
        """
        Set up entry variable traces for validation and event handling
        """

        self._validsettings = True
        self._settype.trace("w", self.val_settings)
        self._setvar.trace("w", self.val_settings)
        self._zoom.trace("w", self.val_settings)
        self._zx_off.trace("w", self.val_settings)
        self._zy_off.trace("w", self.val_settings)
        self._cx_off.trace("w", self.val_settings)
        self._cy_off.trace("w", self.val_settings)
        self._maxiter.trace("w", self.val_settings)
        self._radius.trace("w", self.val_settings)
        self._exponent.trace("w", self.val_settings)
        self._filename.trace("w", self.val_settings)
        self._frames.trace("w", self.val_settings)
        self._zoominc.trace("w", self.val_settings)

    def val_settings(self, *args, **kwargs):
        """
        Validate all user-entered settings.
        (A personal choice but I find this user experience more intuitive
        than the standard validatecommand method for Entry widgets)
        """

        self._validsettings = True
        self.__app.set_status("")

        if self.is_float(self.ent_zoom.get()) and self._zoom.get() > 0:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_zoom, flg)

        if self.is_float(self.ent_zx_off.get()):
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_zx_off, flg)

        if self.is_float(self.ent_zy_off.get()):
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_zy_off, flg)

        if self.is_float(self.ent_cx_off.get()):
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_cx_off, flg)

        if self.is_float(self.ent_cy_off.get()):
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_cy_off, flg)

        if self.is_integer(self.ent_maxiter.get()) and self._maxiter.get() > 0:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_maxiter, flg)

        if self.is_float(self.ent_radius.get()) and self._radius.get() > 0:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_radius, flg)

        if self.is_float(self.ent_exp.get()) and self._exponent.get() >= 1:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_exp, flg)

        if self.is_integer(self.ent_frames.get()) and self._frames.get() > 0:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_frames, flg)

        if self.is_float(self.ent_zoominc.get()) and self._zoominc.get() > 0:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_zoominc, flg)

        if self.is_filename(self.ent_save.get()):
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.ent_save, flg)

        if self.spn_settype.get() == MODES[MANDELBROT]:
            self.btn_autospin.config(state=DISABLED)
            self.ent_cx_off.config(state=DISABLED)
            self.ent_cy_off.config(state=DISABLED)
            self._cx_off.set(0)
            self._cy_off.set(0)
            flg = GOOD
        elif self.spn_settype.get() == MODES[JULIA]:
            self.btn_autospin.config(state=NORMAL)
            self.ent_cx_off.config(state=NORMAL)
            self.ent_cy_off.config(state=NORMAL)
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.spn_settype, flg)

        if self.spn_setvar.get() in VARIANTS:
            flg = GOOD
        else:
            flg = BAD
        self.flag_entry(self.spn_setvar, flg)

    def flag_entry(self, ent, flag):
        """
        Flag entry field as valid or invalid and set global validity status flag.
        This flag is used throughout to determine whether functions can proceed
        or not.
        """

        ent.config(bg=flag)
        if flag == BAD:
            self._validsettings = False
            self.__app.set_status(VALERROR, "red")

    def is_float(self, flag):
        """
        Validate if entry is a float.
        """

        try:
            float(flag)
            return True
        except ValueError:
            return False

    def is_integer(self, flag):
        """
        Validate if entry is a positive integer.
        """

        try:
            int(flag)
            if int(flag) > 0:
                return True
            return False
        except ValueError:
            return False

    def is_filename(self, flag):
        """
        Validate if entry represents a valid filename using a regexp.
        """

        return match(r"^[\w\-. ]+$", flag) and flag != ""

    def reset(self):
        """
        Reset settings to defaults.
        """

        self._settype.set(MODES[0])
        self._setvar.set(VARIANTS[0])
        self._zoom.set(0.75)
        self._zx_off.set(-0.5)
        self._zy_off.set(0.0)
        self._cx_off.set(0.0)
        self._cy_off.set(0.0)
        self._maxiter.set(128)
        self._radius.set(2)
        self._exponent.set(2)
        self._frames.set(10)
        self._zoominc.set(2.0)
        self._autoiter.set(1)
        self._autosave.set(0)
        self._theme.set("Default")
        self._filename.set("image0")
        self._shift.set(0)
        self.set_sel_theme()

        self.__app.set_status(SETINITTXT)

    def get_sel_theme(self, *args, **kwargs):
        """
        Get selected theme from listbox and set global variable.
        """

        idx = self.lbx_theme.curselection()
        if idx == "":
            idx = 0
        self._theme.set(self.lbx_theme.get(idx))

    def set_sel_theme(self):
        """
        Lookup index of selected theme and highlight that listbox position.
        NB: this requires 'exportselection=False' option to be set on listbox to
        work properly.
        """

        for idx, theme in enumerate(THEMES):
            if theme == self._theme.get():
                self.lbx_theme.activate(idx)
                self.lbx_theme.see(idx)
                break

    def get_settings(self):
        """
        Return all current settings as a dict.
        """

        if not self._validsettings:
            settings = {"valid": self._validsettings}
            return settings

        settings = {
            "settype": self._settype.get(),
            "setvar": self._setvar.get(),
            "zoom": self._zoom.get(),
            "zxoffset": self._zx_off.get(),
            "zyoffset": self._zy_off.get(),
            "cxoffset": self._cx_off.get(),
            "cyoffset": self._cy_off.get(),
            "maxiter": self._maxiter.get(),
            "autoiter": self._autoiter.get(),
            "autosave": self._autosave.get(),
            "radius": self._radius.get(),
            "exponent": self._exponent.get(),
            "theme": self._theme.get(),
            "shift": self._shift.get(),
            "filepath": self._filepath,
            "filename": self._filename.get(),
            "frames": self._frames.get(),
            "zoominc": self._zoominc.get(),
            "valid": self._validsettings,
        }
        return settings

    def update_settings(self, **kwargs):
        """
        Update settings from keyword parms.
        """

        if "settype" in kwargs:
            self._settype.set(kwargs["settype"])
        if "setvar" in kwargs:
            self._setvar.set(kwargs["setvar"])
        if "zoom" in kwargs:
            self._zoom.set(kwargs["zoom"])
        if "zxoffset" in kwargs:
            self._zx_off.set(kwargs["zxoffset"])
        if "zyoffset" in kwargs:
            self._zy_off.set(kwargs["zyoffset"])
        if "cxoffset" in kwargs:
            self._cx_off.set(kwargs["cxoffset"])
        if "cyoffset" in kwargs:
            self._cy_off.set(kwargs["cyoffset"])
        if "maxiter" in kwargs:
            self._maxiter.set(kwargs["maxiter"])
        if "autoiter" in kwargs:
            self._autoiter.set(kwargs["autoiter"])
        if "autosave" in kwargs:
            self._autosave.set(kwargs["autosave"])
        if "radius" in kwargs:
            self._radius.set(kwargs["radius"])
        if "exponent" in kwargs:
            self._exponent.set(kwargs["exponent"])
        if "filepath" in kwargs:
            self._filepath.set(kwargs["filepath"])
        if "filename" in kwargs:
            self._filename.set(kwargs["filename"])
        if "frames" in kwargs:
            self._frames.set(kwargs["frames"])
        if "zoominc" in kwargs:
            self._zoominc.set(kwargs["zoominc"])
        if "theme" in kwargs:
            self._theme.set(kwargs["theme"])
            self.set_sel_theme()
        if "shift" in kwargs:
            self._shift.set(kwargs["shift"])

    def set_filepath(self):
        """
        Sets filepath for saved files for the duration of this session.
        """

        default = os.getcwd()  # Default _filepath is current working directory
        if self._filepath is None:
            self._filepath = filedialog.askdirectory(
                title=SAVETITLE, initialdir=default, mustexist=True
            )
            if self._filepath == "":
                self._filepath = None  # User cancelled

        return self._filepath

    def save_image(self):
        """
        Save image as PNG file to selected filepath and automatically increment default
        image name. NB: currently this will overwrite any existing file of the same
        name without warning.
        """

        # Bug out if the settings are invalid
        settings = self.__app.frm_settings.get_settings()
        if not settings["valid"]:
            return

        # Check if image has been created
        image = self.__app.frm_fractal.mandelbrot.get_image()
        if image is None:
            self.__app.set_status(NOIMGERROR, "red")
            return

        # Set _filename and path
        if self.set_filepath() is None:  # User cancelled
            return
        fname = self._filename.get()
        fqname = self._filepath + "/" + fname

        # Save the image along with its metadata
        try:
            # image.write(fqname + ".png", format="png")
            image.save(fqname + ".png", format="png")
            self.save_metadata()
        except OSError:
            self.__app.set_status(SAVEERROR, "red")
            self._filepath = None
            return

        self._image_num += 1
        self._filename.set(self._image_name + str(self._image_num))
        self.__app.set_status(IMGSAVETXT + fqname + ".png", "green")

        # Return focus to image frame
        self.__app.frm_fractal.focus_set()

    def save_metadata(self):
        """
        Save json file containing meta data associated with image,
        allowing it to be imported and reproduced.
        """

        if self._filepath is None:
            if self.set_filepath() is None:  # User cancelled
                return

        fname = self._filename.get()
        fqname = self._filepath + "/" + fname
        filename = fqname + ".json"
        createtime = strftime("%b %d %Y %H:%M:%S %Z", gmtime())

        jsondata = {
            MODULENAME: {
                "filename": fqname + ".png",
                "created": createtime,
                "settype": self._settype.get(),
                "setvar": self._setvar.get(),
                "zoom": self._zoom.get(),
                "zoominc": self._zoominc.get(),
                "frames": self._frames.get(),
                "escradius": self._radius.get(),
                "exponent": self._exponent.get(),
                "maxiter": self._maxiter.get(),
                "zxoffset": self._zx_off.get(),
                "zyoffset": self._zy_off.get(),
                "cxoffset": self._cx_off.get(),
                "cyoffset": self._cy_off.get(),
                "theme": self._theme.get(),
                "shift": self._shift.get(),
            }
        }

        try:
            with open(filename, "w", encoding="utf-8") as outfile:
                dump(jsondata, outfile)
        except OSError:
            self.__app.set_status(METASAVEERROR, "red")
            self._filepath = None

        # Return focus to image frame
        self.__app.frm_fractal.focus_set()

    def import_metadata(self):
        """
        Update settings from imported json metadata file.
        """

        # Select and read file
        try:
            default = os.getcwd()
            filepath = filedialog.askopenfilename(
                initialdir=default,
                title=SELTITLE,
                filetypes=(("json files", "*.json"), ("all files", "*.*")),
            )
            if filepath == "":  # User cancelled
                return
            with open(filepath, "r", encoding="utf-8") as infile:
                jsondata = infile.read()
        except OSError:
            self.__app.set_status(OPENFILEERROR, "red")
            return

        # Parse file
        try:
            settings = loads(jsondata).get(MODULENAME)

            # Set plot parameters
            self._settype.set(settings.get("settype", "Mandelbrot"))
            self._setvar.set(settings.get("setvar", "Standard"))
            self._zoom.set(settings.get("zoom", 1))
            self._zoominc.set(settings.get("zoominc", 2.0))
            self._frames.set(settings.get("frames", 10))
            self._radius.set(settings.get("escradius", 2.0))
            self._exponent.set(settings.get("exponent", 2))
            self._maxiter.set(settings.get("maxiter", 256))
            self._zx_off.set(settings.get("zxoffset", 0))
            self._zy_off.set(settings.get("zyoffset", 0))
            self._cx_off.set(settings.get("cxoffset", 0))
            self._cy_off.set(settings.get("cyoffset", 0))
            self._theme.set(settings.get("theme", "Default"))
            self._shift.set(settings.get("shift", 0))

        except OSError:
            self.__app.set_status(BADJSONERROR, "red")
            return

        self.set_sel_theme()

        fbase = os.path.basename(filepath)
        filename, fileext = os.path.splitext(fbase)
        self.__app.set_status(
            "Metadata file " + filename + fileext + METAPROMPTTXT, "green"
        )

        # Return focus to image frame
        self.__app.frm_fractal.focus_set()



"""
Mandelbrot & Julia set class for tkinter application.

This handles the computation of the Mandelbrot or Julia set as a numpy rgb array
which is loaded into a ImageTk.PhotoImage (and ultimately into a Tk.Canvas
for viewing in the GUI).

NB: use of Numba @jit decorators and pranges to improve performance.
For more information refer to http://numba.pydata.org.

Created on 29 Mar 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=invalid-name

from math import log, sin, sqrt, pi, floor, ceil

from PIL import Image
from numba import jit, prange
import numpy as np

from colormaps.cet_colormap import cet_CBC1, cet_CBTC1, cet_C1, cet_C4s, BlueBrown16
from colormaps.landscape256_colormap import landscape256
from colormaps.metallic256_colormap import metallic256
from colormaps.pastels256_colormap import pastels256
from colormaps.tropical_colormap import tropical16, tropical256
from colormaps.twilight256_colormap import twilight256
from colormaps.twilights512_colormap import twilights512
from colormaps.hsv256_colormap import hsv256

PERIODCHECK = True  # Turn periodicity check optimisation on/off
MANDELBROT = 0
JULIA = 1
STANDARD = 0
BURNINGSHIP = 1
TRICORN = 2
MODES = ("Mandelbrot", "Julia")
VARIANTS = ("Standard", "BurningShip", "Tricorn")
THEMES = [
    "Default",
    "BlueBrown16",
    "Tropical16",
    "Tropical256",
    "Pastels256",
    "Metallic256",
    "Twilight256",
    "Twilights512",
    "Landscape256",
    "Colorcet_CET_C1",
    "Colorcet_CET_CBC1",
    "Colorcet_CET_CBTC1",
    "Colorcet_CET_C4s",
    "HSV256",
    "Monochrome",
    "BasicGrayscale",
    "BasicHue",
    "NormalizedHue",
    "SqrtHue",
    "LogHue",
    "SinHue",
    "SinSqrtHue",
    "BandedRGB",
]


@jit(nopython=True, parallel=True, cache=True)
def plot(
    imagemap,
    settype,
    setvar,
    width,
    height,
    zoom,
    radius,
    exponent,
    zxoff,
    zyoff,
    maxiter,
    theme,
    shift,
    cxoff,
    cyoff,
):
    """
    Plots selected fractal type in the numpy rgb array 'imagemap'
    """

    # For each pixel in array
    for x_axis in prange(width):  # pylint: disable=not-an-iterable
        for y_axis in prange(height):  # pylint: disable=not-an-iterable
            # Invoke core algorithm to calculate escape scalars
            i, za = fractal(
                settype,
                setvar,
                width,
                height,
                x_axis,
                y_axis,
                zxoff,
                zyoff,
                zoom,
                maxiter,
                radius,
                exponent,
                cxoff,
                cyoff,
            )

            # Look up color for these escape scalars and set pixel in imagemap
            imagemap[y_axis, x_axis] = get_color(i, za, radius, maxiter, theme, shift)


@jit(nopython=True, cache=True)
def fractal(
    settype,
    setvar,
    width,
    height,
    x_axis,
    y_axis,
    zxoff,
    zyoff,
    zoom,
    maxiter,
    radius,
    exponent,
    cxoff,
    cyoff,
):
    """
    Calculates fractal escape scalars i, za for each image pixel
    and returns them for use in color rendering routines.
    """

    zx_coord, zy_coord = ptoc(width, height, x_axis, y_axis, zxoff, zyoff, zoom)
    lastz = complex(0, 0)
    per = 0
    i = 0

    z = complex(zx_coord, zy_coord)
    if settype == JULIA:  # Julia or variant
        c = complex(cxoff, cyoff)
    else:  # Mandelbrot or variant
        c = z

    for i in prange(maxiter + 1):  # pylint: disable=not-an-iterable
        # Iterate till the value z is outside the escape radius.
        if setvar == BURNINGSHIP:
            z = complex(abs(z.real), -abs(z.imag))
        if setvar == TRICORN:
            z = z.conjugate()
        z = z**exponent + c

        # Optimisation - periodicity check speeds
        # up processing of points within set
        if PERIODCHECK:
            if z == lastz:
                i = maxiter
                break
            per += 1
            if per > 20:
                per = 0
                lastz = z
        # ... end of optimisation

        if abs(z) > radius**2:
            break

    return i, abs(z)  # i, za


@jit(nopython=True, cache=True)
def ptoc(width, height, x, y, zxoff, zyoff, zoom):
    """
    Converts actual pixel coordinates to complex space coordinates
    (zxoff, zyoff are always the complex offsets).
    """

    zx_coord = zxoff + ((width / height) * (x - width / 2) / (zoom * width / 2))
    zy_coord = zyoff + (-1 * (y - height / 2) / (zoom * height / 2))
    return zx_coord, zy_coord


@jit(nopython=True, cache=True)
def ctop(width, height, zx_coord, zy_coord, zxoff, zyoff, zoom):
    """
    Converts complex space coordinates to actual pixel coordinates
    (zxoff, zyoff are always the complex offsets).
    """

    x_coord = (zx_coord + zxoff) * zoom * width / 2 / (width / height)
    y_coord = (zy_coord + zyoff) * zoom * height / 2
    return x_coord, y_coord


@jit(nopython=True, cache=True)
def get_color(i, za, radius, maxiter, theme, shift):
    """
    Uses escape scalars i, za from the fractal algorithm to drive a variety
    of color rendering algorithms or 'themes'.

    NB: If you want to add more rendering algorithms, this is where to add them,
    but you'll need to ensure they are 'Numba friendly' (i.e. limited to
    Numba-recognised data types and suitably decorated library functions).
    """

    if i == maxiter and theme != "BasicGrayscale":  # Inside Mandelbrot set, so black
        return 0, 0, 0

    if theme == "Default":
        theme = "BlueBrown16"

    if theme == "Monochrome":
        if shift == 0:
            h = 0.0
            s = 0.0
        else:
            h = 0.5 + shift / -200
            s = 1.0
        r, g, b = hsv_to_rgb(h, s, 1.0)
    elif theme == "BasicGrayscale":
        if i == maxiter:
            return 255, 255, 255
        r = 256 * i / maxiter
        b = g = r
    elif theme == "BasicHue":
        h = ((i / maxiter) + (shift / 100)) % 1
        r, g, b = hsv_to_rgb(h, 0.75, 1)
    elif theme == "BandedRGB":
        bands = [0, 32, 96, 192]
        r = bands[(i // 4) % 4]
        g = bands[i % 4]
        b = bands[(i // 16) % 4]
    elif theme == "NormalizedHue":
        h = ((normalize(i, za, radius) / maxiter) + (shift / 100)) % 1
        r, g, b = hsv_to_rgb(h, 0.75, 1)
    elif theme == "SqrtHue":
        h = ((normalize(i, za, radius) / sqrt(maxiter)) + (shift / 100)) % 1
        r, g, b = hsv_to_rgb(h, 0.75, 1)
    elif theme == "LogHue":
        h = ((normalize(i, za, radius) / log(maxiter)) + (shift / 100)) % 1
        r, g, b = hsv_to_rgb(h, 0.75, 1)
    elif theme == "SinHue":
        h = normalize(i, za, radius) * sin(((shift + 1) / 100) * pi / 2)
        r, g, b = hsv_to_rgb(h, 0.75, 1)
    elif theme == "SinSqrtHue":
        steps = 1 + shift / 100
        h = 1 - (sin((normalize(i, za, radius) / sqrt(maxiter) * steps) + 1) / 2)
        r, g, b = hsv_to_rgb(h, 0.75, 1)
    else:  # Indexed colormap arrays
        r, g, b = sel_colormap(i, za, radius, shift, theme)
    return r, g, b


@jit(nopython=True, cache=True)
def sel_colormap(i, za, radius, shift, theme):
    """
    Select from indexed colormap theme
    """

    if theme == "Colorcet_CET_CBC1":
        r, g, b = get_colormap(i, za, radius, shift, cet_CBC1)
    elif theme == "Colorcet_CET_CBTC1":
        r, g, b = get_colormap(i, za, radius, shift, cet_CBTC1)
    elif theme == "Colorcet_CET_C1":
        r, g, b = get_colormap(i, za, radius, shift, cet_C1)
    elif theme == "Colorcet_CET_C4s":
        r, g, b = get_colormap(i, za, radius, shift, cet_C4s)
    elif theme == "BlueBrown16":
        r, g, b = get_colormap(i, za, radius, shift, BlueBrown16)
    elif theme == "Tropical16":
        r, g, b = get_colormap(i, za, radius, shift, tropical16)
    elif theme == "Tropical256":
        r, g, b = get_colormap(i, za, radius, shift, tropical256)
    elif theme == "Pastels256":
        r, g, b = get_colormap(i, za, radius, shift, pastels256)
    elif theme == "Metallic256":
        r, g, b = get_colormap(i, za, radius, shift, metallic256)
    elif theme == "Twilight256":
        r, g, b = get_colormap(i, za, radius, shift, twilight256)
    elif theme == "Twilights512":
        r, g, b = get_colormap(i, za, radius, shift, twilights512)
    elif theme == "Landscape256":
        r, g, b = get_colormap(i, za, radius, shift, landscape256)
    elif theme == "HSV256":
        r, g, b = get_colormap(i, za, radius, shift, hsv256)

    return r, g, b


@jit(nopython=True, cache=True)
def get_colormap(i, za, radius, shift, colmap):
    """
    Get pixel color from colormap
    """

    ni = normalize(i, za, radius)  # normalised iteration count
    sh = ceil(shift * len(colmap) / 100)  # palette shift
    col1 = colmap[(floor(ni) + sh) % len(colmap)]
    col2 = colmap[(floor(ni) + sh + 1) % len(colmap)]
    return interpolate(col1, col2, ni)


@jit(nopython=True, cache=True)
def normalize(i, za, radius):
    """
    Normalize iteration count from escape scalars to produce
    smoother color gradients.
    """

    lzn = log(za**2) / 2
    nu = log(lzn / log(radius)) / log(2)
    return i + 1 - nu


@jit(nopython=True, cache=True)
def interpolate(col1, col2, ni):
    """
    Linear interpolation between two adjacent colors in color palette.
    """

    f = ni % 1  # fractional part of ni
    r = (col2[0] - col1[0]) * f + col1[0]
    g = (col2[1] - col1[1]) * f + col1[1]
    b = (col2[2] - col1[2]) * f + col1[2]
    return [r, g, b]


@jit(nopython=True, cache=True)
def hsv_to_rgb(h, s, v):
    """
    Convert HSV values (in range 0-1) to RGB (in range 0-255).
    """

    v = int(v * 255)
    if s == 0.0:
        return v, v, v
    i = int(h * 6.0)
    f = (h * 6.0) - i
    p = int(v * (1.0 - s))
    q = int(v * (1.0 - s * f))
    t = int(v * (1.0 - s * (1.0 - f)))
    i %= 6
    if i == 0:
        r, g, b = v, t, p
    if i == 1:
        r, g, b = q, v, p
    if i == 2:
        r, g, b = p, v, t
    if i == 3:
        r, g, b = p, q, v
    if i == 4:
        r, g, b = t, p, v
    if i == 5:
        r, g, b = v, p, q

    return r, g, b


class Mandelbrot:
    """
    Main computation and imaging class.
    """

    def __init__(self, master):
        """
        Constructor
        """

        self.__master = master
        self._kill = False
        self._image = None

    def plot_image(
        self,
        settype,
        setvar,
        width,
        height,
        zoom,
        radius,
        exp,
        zxoff,
        zyoff,
        maxiter,
        theme,
        shift,
        cxoff,
        cyoff,
    ):
        """
        Creates empty numpy rgb array, passes it to fractal calculation routine for
        populating, then loads populated array into an ImageTk.PhotoImage.
        """

        self._kill = False
        imagemap = np.zeros((height, width, 3), dtype=np.uint8)
        plot(
            imagemap,
            settype,
            setvar,
            width,
            height,
            zoom,
            radius,
            exp,
            zxoff,
            zyoff,
            maxiter,
            theme,
            shift,
            cxoff,
            cyoff,
        )
        self._image = Image.fromarray(imagemap, "RGB")

    def get_image(self):
        """
        Return populated PhotoImage for display or saving.
        """

        return self._image

    def get_cancel(self):
        """
        Return kill flag.
        """

        return self._kill

    def cancel_plot(self):
        """
        Cancel in-flight plot operation (plot is normally so quick this is
        largely redundant).
        """

        self._kill = True



"""
ENGLISH language string literals for PyMandel tkinter application

Created on 22 Apr 2020

@author: semuadmin
"""
# pylint: disable=line-too-long

WIKIURL = "https://en.wikipedia.org/wiki/Mandelbrot_set"
GITHUBURL = "https://github.com/semuconsulting/PyMandel"
CETURL = "https://github.com/holoviz/colorcet/blob/master/LICENSE.txt"
MODULENAME = "pymandel"

COPYRIGHTTXT = "\u00A9 SEMU Consulting 2020\nGPLv3 License. All Rights Reserved"

COLORCETTXT = "HoloViz Colorcet color maps available under\nCreative Commons Attribution (CC_BY) license"

INTROTXT = (
    "Welcome to PyMandel! Use mouse wheel or left-click to zoom, right-click to center."
)

HELPTXT = (
    "Enter settings manually (or import them from a metadata file) and click PLOT to create a fractal image with the specified parameters.\n\n"
    + "Mouse wheel - zoom in and out at the current cursor location"
    + " (there may be some lag at high resolutions).\n"
    + "Left-click, drag and release - zoom into a drawn rectangular area.\n"
    + "Left-click momentarily - zoom in at the cursor location by the Zoom Increment amount.\n"
    + "Shift & Left-click - zoom out.\n"
    + "Alt-L or Ctrl-L & Left-click - switch to Julia mode and plot the Julia set corresponding to that cursor (cx, cy offset) location.\n"
    + "Right-click - centre the image at the cursor location.\n"
    + "Press Left \u25C0 or Right \u25B6 arrow keys in Julia mode to rotate the Julia Set about its origin.\n\n"
    + "PLOT button - plot the image using current settings.\n"
    + "Cancel button - cancel the current plot operation.\n"
    + "Reset button - reset the parameters to the default values.\n"
    + "Save button - save the currently displayed image as a .png file along with its associated metadata as a .json file.\n\n"
    + "Zoom button - automatically create a sequence of zooming images.\n"
    + "Spin Button - automatically create a sequence of 'spinning' Julia images.\n\n"
    + "File..Export Settings - export current settings (metadata).\n"
    + "File..Import Settings - import previously saved metadata.\n"
    + "Options..Hide/Show Settings - toggles the Settings Panel on or off.\n"
    + "Options..Hide/Show Status - toggles the Status Bar on or off.\n"
    + "Help..Howto - display this How To dialog.\n"
    + "Help..About - display About dialog."
)

ABOUTTXT = (
    "PyMandel is a free, open-source GUI application written entirely in Python and tkinter with Numba performance enhancements.\n\n"
    + "Instructions and source code are available on Github at the link below."
)

# Message text
JITTXT = "FIRST TIME USE ONLY: Please wait for JIT Compilation and Caching"
SETINITTXT = "Settings initialised"
VALERROR = "ERROR! Please correct highlighted entries"
SAVEERROR = "ERROR! File could not be saved to specified directory"
METASAVEERROR = "ERROR! Metadata file could not be saved to specified directory"
NOIMGERROR = "ERROR! An image must be created before saving"
OPENFILEERROR = "ERROR! File could not be opened"
BADJSONERROR = "ERROR! Invalid metadata file"
SAVETITLE = "Select Save Directory"
SELTITLE = "Select File for Import"
METAPROMPTTXT = " imported, click PLOT to proceed"
IMGSAVETXT = "Image saved as "
COMPLETETXT = "Operation completed in "
INPROGTXT = "Operation in progress..."
OPCANTXT = "Operation Cancelled"
COORDTXT = "Coordinates:"
COORDPOLTXT = "Polar coords:"
FRMTXT = "Frame"
FRMSTXT = "Frames"

# Menu text
MENUFILE = "File"
MENUOPTIONS = "Options"
MENUSAVE = "Save Image"
MENUEXPORT = "Export Settings"
MENUIMPORT = "Import Settings"
MENUEXIT = "Exit"
MENUPLOT = "Plot Image"
MENUZOOM = "Plot Zoom Animation"
MENUSPIN = "Plot Julia Spin Animation"
MENUCAN = "Cancel"
MENURST = "Reset"
MENUHIDESE = "Hide Settings"
MENUSHOWSE = "Show Settings"
MENUHIDESB = "Hide Status Bar"
MENUSHOWSB = "Show Status Bar"
MENUHIDEAX = "Hide Axes"
MENUSHOWAX = "Show Axes"
MENUHOWTO = "How To"
MENUABOUT = "About"
MENUHELP = "Help"

# Button text
BTNPLOT = "PLOT"
BTNSAVE = "Save"
BTNCAN = "Cancel"
BTNRST = "Reset"
BTNZOOM = "Zoom"
BTNSPIN = "Spin"

# Label text
LBLCTL = "Controls"
LBLSET = "Settings"
LBLMODE = "Mode"
LBLVAR = "Variant"
LBLAUTO = "Animate:"
LBLTHEME = "Color\nTheme"
LBLEXP = "Exponent"
LBLRAD = "Escape\nRadius"
LBLSHIFT = "Theme\nShift"
LBLITER = "Max\nIterations"
LBLZOOM = "Zoom"
LBLZOOMINC = "Zoom\nIncrement"
LBLZXOFF = "ZX Offset"
LBLZYOFF = "ZY Offset"
LBLCX = "Julia CX"
LBLCY = "Julia CY"

# Dialog text
DLGABOUT = "About PyMandel"
DLGHOWTO = "How To Use PyMandel"



#!/usr/bin/env python
"""
Mandelbrot generator

A command line utility which invokes the Mandelbrot class methods to create a sequence
of Mandelbrot images as .png files. It is the command line equivalent of the GUI
Animate function, but can also be used to generate single, very high resolution images.

Can also import settings from a previously saved metadata file.

Created on 29 Mar 2020

:author: semuadmin
:copyright: SEMU Consulting © 2020
:license: GPL3

This file is part of PyMandel.

PyMandel is free software: you can redistribute it and/or modify it under the terms of the
GNU General Public License as published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

PyMandel is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with PyMandel.
If not, see <https://www.gnu.org/licenses/>.
"""

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, SUPPRESS
from json import loads
from math import log, sqrt
import sys
from time import time

from pymandel.mandelbrot import Mandelbrot, MANDELBROT, JULIA
from pymandel.strings import MODULENAME
from pymandel._version import __version__ as VERSION

sys.path.append("pymandel")
sys.path.append("colormaps")

EPILOG = (
    "© 2021 SEMU Consulting GPLv3 license - https://github.com/semuconsulting/PyMandel/"
)


class BatchMandelbrot:
    """
    Main class of command line utility
    """

    def __init__(self, **kwargs):
        """
        Command line utility class
        """

        print("Parameters passed: ", kwargs)
        self._setmode = kwargs.get("settype", "Mandelbrot")
        if self._setmode == "Julia":
            self._settype = JULIA
        else:
            self._settype = MANDELBROT
        self._setvar = kwargs.get("setvar", "Standard")
        self._width = int(kwargs.get("width", 1920))
        self._height = int(kwargs.get("height", 1080))
        self._radius = int(kwargs.get("escradius", 2))
        self._exponent = int(kwargs.get("exponent", 2))
        self._zx_off = float(kwargs.get("zxoffset", -0.5))
        self._zy_off = float(kwargs.get("zyoffset", 0.0))
        self._cx_off = float(kwargs.get("cxoffset", 0.0))
        self._cy_off = float(kwargs.get("cyoffset", 0.0))
        self._filepath = kwargs.get("filepath", ".")
        self._filename = kwargs.get("filename", "image")
        self._frames = int(kwargs.get("frames", 1))
        self._zoominc = float(kwargs.get("zoominc", 1.2))
        self._theme = kwargs.get("theme", "Default")
        self._shift = int(kwargs.get("shift", 0))
        self._startframe = int(kwargs.get("startframe", 1))
        if self._startframe > self._frames:
            self._startframe = self._frames
        self._zoom = float(kwargs.get("zoom", 0.75))
        self._startzoom = float(
            kwargs.get(
                "startzoom", self._zoom * pow(self._zoominc, self._startframe - 1)
            )
        )
        self._zoom = self._startzoom
        self._maxiter = int(
            kwargs.get("maxiter", abs(1000 * log(1 / sqrt(self._zoom))))
        )

        self._importfile = kwargs.get("import", "")
        if self._importfile != "":
            if not self.import_metadata(self._importfile):
                return

        start = time()

        i = self.animate()

        end = time()
        print(f"Sequence of {i+1} frames took {round(end - start, 2)} secs")

    def animate(self) -> int:
        """
        Generates and saves a series of frames at a specific point and zoom increment.
        """

        i = 0
        try:
            self.mandelbrot = Mandelbrot(self)
            self.mandelbrot.cancel_plot()  # Cancel any in-flight plot

            for i in range(self._frames):
                self._currframe = i
                fqname = f"{self._filepath}/{self._filename}_{(i + 1):03d}"
                print(f"Creating file {fqname} ...")

                self.mandelbrot.plot_image(
                    self._settype,
                    self._setvar,
                    self._width,
                    self._height,
                    self._zoom,
                    self._radius,
                    self._exponent,
                    self._zx_off,
                    self._zy_off,
                    self._maxiter,
                    self._theme,
                    self._shift,
                    self._cx_off,
                    self._cy_off,
                )
                image = self.mandelbrot.get_image()

                try:
                    image.save(f"{fqname}.png", format="png")
                except OSError:
                    print(f"ERROR! File {fqname} could not be saved to specified path")
                    return i

                self._zoom = self._zoom * self._zoominc
                self._maxiter = self.get_autoiter(self._zoom)
        except KeyboardInterrupt:
            print("Animation interrupted by user")

        print("Animation complete")
        return i

    def import_metadata(self, filepath):
        """
        Import settings from json metadata file.
        """

        # Read file
        try:
            with open(filepath, "r", encoding="utf-8") as infile:
                jsondata = infile.read()
        except OSError:
            print("ERROR! Unable to read import file")
            return False

        # Parse file
        settings = loads(jsondata)
        self._settype = settings[MODULENAME]["settype"]
        self._setvar = settings[MODULENAME]["setvar"]
        self._zoom = float(settings[MODULENAME]["zoom"])
        self._radius = float(settings[MODULENAME]["escradius"])
        self._exponent = int(settings[MODULENAME]["exponent"])
        self._maxiter = int(settings[MODULENAME]["maxiter"])
        self._zx_off = float(settings[MODULENAME]["zxoffset"])
        self._zy_off = float(settings[MODULENAME]["zyoffset"])
        self._cx_off = float(settings[MODULENAME]["cxoffset"])
        self._cy_off = float(settings[MODULENAME]["cyoffset"])
        self._theme = settings[MODULENAME]["theme"]
        self._shift = int(settings[MODULENAME]["shift"])
        self._frames = int(settings[MODULENAME]["frames"])

        return True

    def get_autoiter(self, zoom):
        """
        Arbitrary algorithm to derive 'optimal' max iterations from zoom level.
        """

        if self._settype == JULIA:
            miniter = 500
        else:
            miniter = 100
        maxiter = max(miniter, int(abs(1000 * log(1 / sqrt(zoom)))))
        return maxiter


def main():
    """Entry point for CLI."""

    arp = ArgumentParser(
        epilog=EPILOG,
        formatter_class=ArgumentDefaultsHelpFormatter,
        argument_default=SUPPRESS,
    )
    arp.add_argument("-V", "--version", action="version", version="%(prog)s " + VERSION)
    arp.add_argument(
        "--settype",
        help="Set Type",
        choices=["Mandelbrot", "Julia"],
        default="Mandelbrot",
    )
    arp.add_argument(
        "--setvar",
        help="Set Variant",
        choices=["Standard", "BurningShip", "Tricorn"],
        default="Standard",
    )
    arp.add_argument(
        "--width", help="Width of the image(s) in pixels", type=int, default=1920
    )
    arp.add_argument(
        "--height", help="Height of the image(s) in pixels", type=int, default=1080
    )
    arp.add_argument("--zoom", help="Initial zoom level", type=float, default=0.75)
    arp.add_argument(
        "--maxiter", help="Initial maximum iterations", type=int, default=256
    )
    arp.add_argument("--zxoffset", help="X (Re) axis offset", type=float, default=-0.5)
    arp.add_argument("--zyoffset", help="Y (Im) axis offset", type=float, default=0.0)
    arp.add_argument(
        "--cxoffset", help="CX (Re) axis offset for Julia sets", type=float, default=0.0
    )
    arp.add_argument(
        "--cyoffset", help="CY (Im) axis offset for Julia sets", type=float, default=0.0
    )
    arp.add_argument("--escradius", help="Escape radius", type=float, default=2.0)
    arp.add_argument("--exponent", help="Iteration exponent", type=int, default=2)
    arp.add_argument("--frames", help="Number of frames to create", type=int, default=1)
    arp.add_argument("--startframe", help="Starting frame number", type=int, default=1)
    arp.add_argument(
        "--zoominc", help="Zoom increment between frames", type=float, default=1.2
    )
    arp.add_argument("--theme", help="Color rendering theme", default="Default")
    arp.add_argument("--shift", help="Color theme shift", type=int, default=0)
    arp.add_argument("--filepath", help="Path for saved files", default=".")
    arp.add_argument("--filename", help="Name prefix for saved files", default="frame")
    arp.add_argument(
        "--import", help="Fully qualified path to a previously saved metadata file"
    )

    kwargs = vars(arp.parse_args())
    BatchMandelbrot(**kwargs)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
Command line utility to create numpy color map arrays suitable for use by PyMandel
from image files containing suitable color gradients e.g. created using GIMP's gradient tool.
It currently only handles RGB or RGBA formats.

make_colormap --mapname mymap --input mymap.png --output mymap_colormap.py --levels 256

Created on 24 Apr 2020

:author: semuadmin
:copyright: SEMU Consulting 2020
:license: BSD 3-Clause
"""

import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, SUPPRESS
from PIL import Image
from pymandel._version import __version__ as VERSION

EPILOG = (
    "© 2021 SEMU Consulting GPLv3 license - https://github.com/semuconsulting/PyMandel/"
)


def make_colormap(**kwargs):
    """
    Scan image file and create array of rgb values

    :param kwargs: optional keyword args
    """

    print("Parameters passed: ", kwargs)

    mapname = kwargs.get("mapname", "mymap")
    infile = kwargs.get("input", mapname + ".png")

    try:
        image = Image.open(infile)
    except FileNotFoundError as err:
        print(f"Input file {infile} not found \n{err}")
        return

    mode = image.mode
    if mode not in ("RGB", "RGBA"):
        print(f"Invalid image mode {mode} - must be RGB or RGBA")

    levels = int(kwargs.get("levels", image.width))
    if levels > image.width:
        levels = image.width
    outfile = kwargs.get("output", mapname + str(levels) + "_colormap.py")

    file = open(outfile, "a", encoding="utf-8")
    file.write("from numpy import array\n\n")
    file.write(
        "#****************************************************************************************\n"
        f"# {str(levels)}-level colormap created by make_colormap utility from file {infile}\n"
        "#****************************************************************************************\n"
    )
    file.write(mapname + " = array([ \\\n")

    for x_axis in range(levels):
        pix = int(x_axis * image.width / levels)
        if x_axis == levels - 1:
            end = "]])\n"
        elif x_axis % 4 == 3:
            end = "],\n"
        else:
            end = "],"
        if mode == "RGBA":
            red, grn, blu, alpha = image.getpixel((pix, 0))
        else:
            red, grn, blu = image.getpixel((pix, 0))
        file.write("[" + str(red) + "," + str(grn) + "," + str(blu) + end)

    image.close()
    file.close()

    print(str(levels) + "-level colormap file " + outfile + " created")


def main():
    """CLI entry point."""

    arp = ArgumentParser(
        epilog=EPILOG,
        formatter_class=ArgumentDefaultsHelpFormatter,
        argument_default=SUPPRESS,
    )
    arp.add_argument("-V", "--version", action="version", version="%(prog)s " + VERSION)
    arp.add_argument("--mapname", help="Map name", default="mymap")
    arp.add_argument("--input", help="Input image filename", default="mymap.png")
    arp.add_argument("--levels", help="Color levels", type=int, default=256)
    arp.add_argument("--output", help="Output filename", default="mymap_colormap.py")

    kwargs = vars(arp.parse_args())
    make_colormap(**kwargs)


if __name__ == "__main__":
    sys.exit(main())



"""
Literales de cadena de idioma ESPAÑOL para la aplicación PyMandel tkinter

Creado el 22 abr 2020

@autor: semuadmin
"""
# pylint: disable=line-too-long

WIKIURL = "https://en.wikipedia.org/wiki/Mandelbrot_set"
GITHUBURL = "https://github.com/semuconsulting/PyMandel"
CETURL = "https://github.com/holoviz/colorcet/blob/master/LICENSE.txt"
MODULENAME = "pymandel"

COPYRIGHTTXT = (
    "\u00A9 SEMU Consulting 2020 \nGPLv3 Licencia. Todos los derechos reservados"
)

COLORCETTXT = "Mapas de color HoloViz Colorcet disponibles bajo licencia de\nCreative Commons Attribution (CC_BY)"

INTROTXT = "Bienvenido a PyMandel! Use la rueda del mouse o haga clic izquierdo para hacer zoom, haga clic derecho para centrar."

HELPTXT = (
    "Ingrese la configuración manualmente (o impórtela de un archivo de metadatos) y haga clic en TRAZER para crear una imagen fractal con los parámetros especificados.\n\n"
    + "Use la rueda del mouse para acercar y alejar la ubicación actual del cursor.\n"
    + "Clic izquierdo, arrastre y suelte: amplíe un área rectangular dibujada.\n"
    + "Clic izquierdo - acerca la ubicación del cursor por la cantidad de incremento de zoom.\n"
    + "Shift y Clic izquierdo - alejar.\n"
    + "Presione Izquierda \u25C0 o Derecha \u25B6 en modo Julia para rotar el Conjunto Julia sobre su origen.\n\n"
    + "Botón TRAZER - trazar la imagen con la configuración actual.\n"
    + "Botón Guardar: guarda la imagen que se muestra actualmente como un archivo .png junto con sus metadatos asociados como un archivo .json.\n"
    + "Botón Cancelar: cancela la operación de trazado actual.\n"
    + "Botón Restablecer: restablece los parámetros a los valores predeterminados.\n\n"
    + "Botón de zoom: crea automáticamente una secuencia de imágenes con zoom.\n"
    + "Botón Girar: crea automáticamente una secuencia de imágenes giratorias de Julia.\n\n"
    + "Archivo..Configuración de exportación: exporta la configuración actual (metadatos).\n"
    + "Archivo..Configuración de importación: importa metadatos guardados previamente.\n"
    + "Opciones..Ocultar / Mostrar configuración: activa o desactiva el Panel de configuración.\n"
    + "Opciones..Ocultar / Mostrar estado: activa o desactiva la barra de estado.\n"
    + "Ayuda..Cómo: mostrar este cuadro de diálogo Cómo.\n"
    + "Ayuda..Acerca de: muestra el cuadro de diálogo Acerca de."
)

ABOUTTXT = (
    "PyMandel es una aplicación GUI gratuita de código abierto escrita completamente en Python y tkinter con mejoras de rendimiento de Numba.\n\n"
    + "Las instrucciones y el código fuente están disponibles en Github en el siguiente enlace."
)

# Texto de Mensaje
JITTXT = "USO POR PRIMERA VEZ SOLO: espere la compilación y el almacenamiento en caché de JIT"
SETINITTXT = "Configuración inicializada"
VALERROR = "ERROR! Corrija las entradas resaltadas"
SAVEERROR = "ERROR! El archivo no se pudo guardar en el directorio especificado"
METASAVEERROR = (
    "ERROR! El archivo de metadatos no se pudo guardar en el directorio especificado"
)
NOIMGERROR = "ERROR! Se debe crear una imagen antes de guardar"
OPENFILEERROR = "ERROR! El archivo no se pudo abrir"
BADJSONERROR = "ERROR! Archivo de metadatos no válido"
SAVETITLE = "Seleccionar Guardar Directorio"
SELTITLE = "Seleccionar archivo para importar"
METAPROMPTTXT = "importado, haga clic en TRAZER para continuar"
IMGSAVETXT = "Imagen guardada como"
COMPLETETXT = "Operación completada en"
INPROGTXT = "Operación en progreso ..."
OPCANTXT = "Operación cancelada"
COORDTXT = "Coordenadas:"
COORDPOLTXT = "Cordones polares:"
FRMTXT = "Marco"
FRMSTXT = "Marcos"

# Texto de Menú
MENUFILE = "Archivo"
MENUOPTIONS = "Opciones"
MENUSAVE = "Guardar Imagen"
MENUEXPORT = "Exportar Parámetros"
MENUIMPORT = "Importar Parámetros"
MENUEXIT = "Salir"
MENUPLOT = "Trazar Imagen"
MENUZOOM = "Trazar Animación de zoom"
MENUSPIN = "Trazar Animación de Rotación Julia"
MENUCAN = "Cancelar"
MENURST = "Restablecer"
MENUHIDESE = "Ocultar Parámetros"
MENUSHOWSE = "Mostrar Parámetros"
MENUHIDESB = "Ocultar Barra de Estado"
MENUSHOWSB = "Mostrar Barra de Estado"
MENUHIDEAX = "Ocultar Ejes"
MENUSHOWAX = "Mostrar Ejes"
MENUHOWTO = "Cómo"
MENUABOUT = "Acerca de"
MENUHELP = "Ayuda"

# Texto de Botón
BTNPLOT = "TRAZER"
BTNSAVE = "Guardar"
BTNCAN = "Cancelar"
BTNRST = "Restablecer"
BTNZOOM = "Zoom"
BTNSPIN = "Girar"

# Texto de etiqueta
LBLCTL = "Controles"
LBLSET = "Parámetros"
LBLMODE = "Conjunto tipo"
LBLVAR = "Variant"
LBLAUTO = "Animar:"
LBLTHEME = "Tema"
LBLEXP = "Exponente"
LBLRAD = "Escape\nRadio"
LBLSHIFT = "Tema\nCambiar"
LBLITER = "Max\nIteraciones"
LBLZOOM = "Zoom"
LBLZOOMINC = "Zoom\nIncremento"
LBLZXOFF = "ZX Offset"
LBLZYOFF = "ZY Offset"
LBLCX = "Julia CX"
LBLCY = "Julia CY"

# Texto de diálogo
DLGABOUT = "Acerca de PyMandel"
DLGHOWTO = "Cómo usar PyMandel"
